@@ -2,10 +2,13 @@
 Image and target transform utilities.
 """
 
+from dataclasses import dataclass
 from typing import Any, Callable, List, Tuple, TypeVar
 
 import numpy as np  # type: ignore
 
+from targetran.utils import Interpolation
+
 # This roughly means anything that is ndarray-like.
 T = TypeVar("T", np.ndarray, Any)
 
@@ -32,33 +35,42 @@ def _sanitise(
     return image, bboxes, labels
 
 
+@dataclass
+class _AffineDependency:
+    convert_fn: Callable[..., T]
+    shape_fn: Callable[[T], Tuple[int, ...]]
+    reshape_fn: Callable[[T, Tuple[int, ...]], T]
+    expand_dim_fn: Callable[[T, int], T]
+    squeeze_fn: Callable[[T, int], T]
+    pad_image_fn: Callable[[T, T], T]
+    range_fn: Callable[[int, int, int], T]
+    cast_to_int_fn: Callable[[T], T]
+    round_to_int_fn: Callable[[T], T]
+    repeat_fn: Callable[[T, T], T]
+    tile_fn: Callable[[T, T], T]
+    ones_like_fn: Callable[[T], T]
+    stack_fn: Callable[[List[T], int], T]
+    concat_fn: Callable[[List[T], int], T]
+    matmul_fn: Callable[[T, T], T]
+    clip_fn: Callable[[T, T, T], T]
+    floor_fn: Callable[[T], T]
+    ceil_fn: Callable[[T], T]
+    gather_image_fn: Callable[[T, T], T]
+    copy_fn: Callable[[T], T]
+    max_fn: Callable[[T, int], T]
+    min_fn: Callable[[T, int], T]
+    logical_and_fn: Callable[[T, T], T]
+    boolean_mask_fn: Callable[[T, T], T]
+
+
 def _affine_transform(
         image: T,
         bboxes: T,
         labels: T,
-        convert_fn: Callable[..., T],
-        shape_fn: Callable[[T], Tuple[int, ...]],
-        reshape_fn: Callable[[T, Tuple[int, ...]], T],
-        expand_dim_fn: Callable[[T, int], T],
-        squeeze_fn: Callable[[T, int], T],
-        pad_image_fn: Callable[[T, T], T],
-        range_fn: Callable[[int, int, int], T],
-        round_to_int_fn: Callable[[T], T],
-        repeat_fn: Callable[[T, T], T],
-        tile_fn: Callable[[T, T], T],
-        ones_like_fn: Callable[[T], T],
-        stack_fn: Callable[[List[T], int], T],
-        concat_fn: Callable[[List[T], int], T],
         image_dest_tran_mat: T,
         bboxes_tran_mat: T,
-        matmul_fn: Callable[[T, T], T],
-        clip_fn: Callable[[T, T, T], T],
-        gather_image_fn: Callable[[T, T], T],
-        copy_fn: Callable[[T], T],
-        max_fn: Callable[[T, int], T],
-        min_fn: Callable[[T, int], T],
-        logical_and_fn: Callable[[T, T], T],
-        boolean_mask_fn: Callable[[T, T], T]
+        interpolation: Interpolation,
+        d: _AffineDependency
 ) -> Tuple[T, T, T]:
     """
     image: [h, w, c]
@@ -66,106 +78,138 @@ def _affine_transform(
     labels: [0, 1, 0, ...]
     """
     image, bboxes, labels = _sanitise(
-        image, bboxes, labels, convert_fn, shape_fn, reshape_fn
+        image, bboxes, labels, d.convert_fn, d.shape_fn, d.reshape_fn
     )
 
-    image_shape = shape_fn(image)
+    image_shape = d.shape_fn(image)
     height, width = int(image_shape[0]), int(image_shape[1])
     h_mod, w_mod = height % 2, width % 2
     num_channels = int(image_shape[2])
 
     # Pad image to provide a zero-value pixel frame for clipping use below.
-    pad_offsets = convert_fn([1, 1, 1, 1])
-    image = pad_image_fn(image, pad_offsets)
+    pad_offsets = d.convert_fn([1, 1, 1, 1])
+    image = d.pad_image_fn(image, pad_offsets)
 
     # References:
     # https://www.kaggle.com/cdeotte/rotation-augmentation-gpu-tpu-0-96
 
     # Destination indices. Note that (-foo // 2) != -(foo // 2).
-    row_idxes = repeat_fn(  # Along y-axis, from top to bottom.
-        range_fn(-(height // 2) + 1 - h_mod, height // 2 + 1, 1),
-        round_to_int_fn(convert_fn([width]))
+    row_idxes = d.repeat_fn(  # Along y-axis, from top to bottom.
+        d.range_fn(-(height // 2) + 1 - h_mod, height // 2 + 1, 1),
+        d.round_to_int_fn(d.convert_fn([width]))
     )
-    col_idxes = tile_fn(  # Along x-axis, from left to right.
-        range_fn(-(width // 2) + 1 - w_mod, width // 2 + 1, 1),
-        round_to_int_fn(convert_fn([height]))
+    col_idxes = d.tile_fn(  # Along x-axis, from left to right.
+        d.range_fn(-(width // 2) + 1 - w_mod, width // 2 + 1, 1),
+        d.round_to_int_fn(d.convert_fn([height]))
     )
     # Note the (col, row) -> (x, y) swapping. Last axis is dummy.
-    image_dest_idxes = stack_fn(
-        [col_idxes, row_idxes, ones_like_fn(col_idxes)], 0
+    image_dest_idxes = d.stack_fn(
+        [col_idxes, row_idxes, d.ones_like_fn(col_idxes)], 0
     )
 
-    # Transform image, with clipping.
-    new_image_dest_idxes = matmul_fn(
-        image_dest_tran_mat, convert_fn(image_dest_idxes)
+    # Transform destination indices, with clipping. Note that these are floats.
+    new_image_dest_idxes = d.matmul_fn(
+        image_dest_tran_mat, d.convert_fn(image_dest_idxes)
     )
-    clipped_new_image_dest_idxes = clip_fn(
+    clipped_new_image_dest_idxes = d.clip_fn(
         new_image_dest_idxes[:2],
         # Note the extra idx for the padded frame.
-        convert_fn([
+        d.convert_fn([
             [-(width // 2) - w_mod], [-(height // 2) - h_mod]
         ]),
-        convert_fn([
+        d.convert_fn([
             [width // 2 + 1], [height // 2 + 1]
         ])
     )
 
     # Assigning original pixel values to new positions.
-    image_orig_idxes = concat_fn([
+    image_orig_idxes = d.concat_fn([
         # Rows.
-        clipped_new_image_dest_idxes[1:] + convert_fn(height // 2 + h_mod),
+        clipped_new_image_dest_idxes[1:] + d.convert_fn(height // 2 + h_mod),
         # Columns.
-        clipped_new_image_dest_idxes[:1] + convert_fn(width // 2 + w_mod)
+        clipped_new_image_dest_idxes[:1] + d.convert_fn(width // 2 + w_mod)
     ], 0)
-    image_orig_idxes = round_to_int_fn(image_orig_idxes)
-    values = gather_image_fn(image, image_orig_idxes)
-    new_image = reshape_fn(values, (height, width, num_channels))
+
+    if interpolation == Interpolation.NEAREST:
+        image_rounded_orig_idxes = d.round_to_int_fn(image_orig_idxes)
+        values = d.gather_image_fn(image, image_rounded_orig_idxes)
+
+    elif interpolation == Interpolation.BILINEAR:
+        floor_floor_idxes = d.cast_to_int_fn(d.floor_fn(image_orig_idxes))
+        ceil_ceil_idxes = d.cast_to_int_fn(d.ceil_fn(image_orig_idxes))
+
+        floor_ceil_idxes = d.concat_fn(
+            [floor_floor_idxes[:1, :], ceil_ceil_idxes[1:, :]], 0
+        )
+        ceil_floor_idxes = d.concat_fn(
+            [ceil_ceil_idxes[:1, :], floor_floor_idxes[1:, :]], 0
+        )
+
+        dists = image_orig_idxes - d.convert_fn(floor_floor_idxes)
+        # Reshape needed for broadcasting in the gather step.
+        floor_weights = d.reshape_fn(1.0 - dists, (-1, 2))
+        ceil_weights = 1.0 - floor_weights
+        floor_floor_weights = floor_weights[:, :1] * floor_weights[:, 1:]
+        floor_ceil_weights = floor_weights[:, :1] * ceil_weights[:, 1:]
+        ceil_floor_weights = ceil_weights[:, :1] * floor_weights[:, 1:]
+        ceil_ceil_weights = ceil_weights[:, :1] * ceil_weights[:, 1:]
+
+        values = \
+            d.gather_image_fn(image, floor_floor_idxes) * floor_floor_weights + \
+            d.gather_image_fn(image, floor_ceil_idxes) * floor_ceil_weights + \
+            d.gather_image_fn(image, ceil_floor_idxes) * ceil_floor_weights + \
+            d.gather_image_fn(image, ceil_ceil_idxes) * ceil_ceil_weights
+
+    else:
+        raise ValueError("Undefined interpolation option.")
+
+    new_image = d.reshape_fn(values, (height, width, num_channels))
 
     # Transform bboxes.
     top_left_xs = bboxes[:, :1]
     top_left_ys = bboxes[:, 1:2]
     top_right_xs = bboxes[:, :1] + bboxes[:, 2:3] - 1
     top_right_ys = bboxes[:, 1:2]
-    bottom_left_xs = copy_fn(top_left_xs)
-    bottom_left_ys = copy_fn(top_left_ys + bboxes[:, 3:] - 1)
-    bottom_right_xs = copy_fn(top_right_xs)
-    bottom_right_ys = copy_fn(top_right_ys + bboxes[:, 3:] - 1)
+    bottom_left_xs = d.copy_fn(top_left_xs)
+    bottom_left_ys = d.copy_fn(top_left_ys + bboxes[:, 3:] - 1)
+    bottom_right_xs = d.copy_fn(top_right_xs)
+    bottom_right_ys = d.copy_fn(top_right_ys + bboxes[:, 3:] - 1)
 
-    xs = concat_fn(
-        [top_left_xs - convert_fn(width // 2 - 1 + w_mod),
-         top_right_xs - convert_fn(width // 2 - 1 + w_mod),
-         bottom_left_xs - convert_fn(width // 2 - 1 + w_mod),
-         bottom_right_xs - convert_fn(width // 2 - 1 + w_mod)],
+    xs = d.concat_fn(
+        [top_left_xs - d.convert_fn(width // 2 - 1 + w_mod),
+         top_right_xs - d.convert_fn(width // 2 - 1 + w_mod),
+         bottom_left_xs - d.convert_fn(width // 2 - 1 + w_mod),
+         bottom_right_xs - d.convert_fn(width // 2 - 1 + w_mod)],
         1
     )
-    ys = concat_fn(
-        [top_left_ys - convert_fn(height // 2 - 1 + h_mod),
-         top_right_ys - convert_fn(height // 2 - 1 + h_mod),
-         bottom_left_ys - convert_fn(height // 2 - 1 + h_mod),
-         bottom_right_ys - convert_fn(height // 2 - 1 + h_mod)],
+    ys = d.concat_fn(
+        [top_left_ys - d.convert_fn(height // 2 - 1 + h_mod),
+         top_right_ys - d.convert_fn(height // 2 - 1 + h_mod),
+         bottom_left_ys - d.convert_fn(height // 2 - 1 + h_mod),
+         bottom_right_ys - d.convert_fn(height // 2 - 1 + h_mod)],
         1
     )
-    bboxes_idxes = stack_fn(  # Shape: [num_bboxes, 3, 4].
-        [xs, ys, ones_like_fn(xs)], 1
+    bboxes_idxes = d.stack_fn(  # Shape: [num_bboxes, 3, 4].
+        [xs, ys, d.ones_like_fn(xs)], 1
     )
 
-    tran_bboxes_idxes = matmul_fn(bboxes_tran_mat, bboxes_idxes)
+    tran_bboxes_idxes = d.matmul_fn(bboxes_tran_mat, bboxes_idxes)
 
     # New bboxes, defined as the rectangle enclosing the transformed bboxes.
     tran_xs = tran_bboxes_idxes[:, 0, :]  # Shape: [num_bboxes, 4].
     tran_ys = tran_bboxes_idxes[:, 1, :]
-    max_xs = max_fn(tran_xs, -1)  # Shape: [num_bboxes].
-    max_ys = max_fn(tran_ys, -1)
-    min_xs = min_fn(tran_xs, -1)
-    min_ys = min_fn(tran_ys, -1)
+    max_xs = d.max_fn(tran_xs, -1)  # Shape: [num_bboxes].
+    max_ys = d.max_fn(tran_ys, -1)
+    min_xs = d.min_fn(tran_xs, -1)
+    min_ys = d.min_fn(tran_ys, -1)
 
-    tran_top_left_xs = round_to_int_fn(expand_dim_fn(min_xs, -1))
-    tran_top_left_ys = round_to_int_fn(expand_dim_fn(min_ys, -1))
-    tran_bottom_right_xs = round_to_int_fn(expand_dim_fn(max_xs, -1))
-    tran_bottom_right_ys = round_to_int_fn(expand_dim_fn(max_ys, -1))
+    tran_top_left_xs = d.round_to_int_fn(d.expand_dim_fn(min_xs, -1))
+    tran_top_left_ys = d.round_to_int_fn(d.expand_dim_fn(min_ys, -1))
+    tran_bottom_right_xs = d.round_to_int_fn(d.expand_dim_fn(max_xs, -1))
+    tran_bottom_right_ys = d.round_to_int_fn(d.expand_dim_fn(max_ys, -1))
     new_widths = tran_bottom_right_xs - tran_top_left_xs + 1
     new_heights = tran_bottom_right_ys - tran_top_left_ys + 1
-    tran_bboxes = concat_fn([  # Shape: [num_bboxes, 4].
+    tran_bboxes = d.concat_fn([  # Shape: [num_bboxes, 4].
         tran_top_left_xs, tran_top_left_ys, new_widths, new_heights
     ], -1)
 
@@ -175,28 +219,32 @@ def _affine_transform(
     # Filter new bboxes values.
     xcens = new_xs + new_widths // 2
     ycens = new_ys + new_heights // 2
-    included = squeeze_fn(logical_and_fn(
-        logical_and_fn(xcens >= 0, xcens <= width),
-        logical_and_fn(ycens >= 0, ycens <= height)
+    included = d.squeeze_fn(d.logical_and_fn(
+        d.logical_and_fn(xcens >= 0, xcens <= width),
+        d.logical_and_fn(ycens >= 0, ycens <= height)
     ), -1)
 
     # Clip new bboxes values.
-    xmaxs = clip_fn(
-        convert_fn(new_xs + new_widths), convert_fn(0), convert_fn(width)
+    xmaxs = d.clip_fn(
+        d.convert_fn(new_xs + new_widths), d.convert_fn(0), d.convert_fn(width)
     )
-    ymaxs = clip_fn(
-        convert_fn(new_ys + new_heights), convert_fn(0), convert_fn(height)
+    ymaxs = d.clip_fn(
+        d.convert_fn(new_ys + new_heights), d.convert_fn(0), d.convert_fn(height)
+    )
+    new_xs = d.clip_fn(
+        d.convert_fn(new_xs), d.convert_fn(0), d.convert_fn(width)
+    )
+    new_ys = d.clip_fn(
+        d.convert_fn(new_ys), d.convert_fn(0), d.convert_fn(height)
     )
-    new_xs = clip_fn(convert_fn(new_xs), convert_fn(0), convert_fn(width))
-    new_ys = clip_fn(convert_fn(new_ys), convert_fn(0), convert_fn(height))
     new_widths = xmaxs - new_xs
     new_heights = ymaxs - new_ys
 
-    new_bboxes = concat_fn([new_xs, new_ys, new_widths, new_heights], 1)
-    new_bboxes = convert_fn(boolean_mask_fn(new_bboxes, included))
+    new_bboxes = d.concat_fn([new_xs, new_ys, new_widths, new_heights], 1)
+    new_bboxes = d.convert_fn(d.boolean_mask_fn(new_bboxes, included))
 
     # Filter labels.
-    new_labels = boolean_mask_fn(labels, included)
+    new_labels = d.boolean_mask_fn(labels, included)
 
     return new_image, new_bboxes, new_labels
 
@@ -221,27 +269,8 @@ def _flip_left_right(
         image: T,
         bboxes: T,
         labels: T,
-        convert_fn: Callable[..., T],
-        shape_fn: Callable[[T], Tuple[int, ...]],
-        reshape_fn: Callable[[T, Tuple[int, ...]], T],
-        expand_dim_fn: Callable[[T, int], T],
-        squeeze_fn: Callable[[T, int], T],
-        pad_image_fn: Callable[[T, T], T],
-        range_fn: Callable[[int, int, int], T],
-        round_to_int_fn: Callable[[T], T],
-        repeat_fn: Callable[[T, T], T],
-        tile_fn: Callable[[T, T], T],
-        ones_like_fn: Callable[[T], T],
-        stack_fn: Callable[[List[T], int], T],
-        concat_fn: Callable[[List[T], int], T],
-        matmul_fn: Callable[[T, T], T],
-        clip_fn: Callable[[T, T, T], T],
-        gather_image_fn: Callable[[T, T], T],
-        copy_fn: Callable[[T], T],
-        max_fn: Callable[[T, int], T],
-        min_fn: Callable[[T, int], T],
-        logical_and_fn: Callable[[T, T], T],
-        boolean_mask_fn: Callable[[T, T], T]
+        interpolation: Interpolation,
+        d: _AffineDependency
 ) -> Tuple[T, T, T]:
     """
     image: [h, w, c]
@@ -249,15 +278,11 @@ def _flip_left_right(
     labels: [0, 1, 0, ...]
     """
     image_dest_flip_lr_mat, bboxes_flip_lr_mat = _get_flip_left_right_mats(
-        convert_fn
+        d.convert_fn
     )
     return _affine_transform(
-        image, bboxes, labels, convert_fn, shape_fn, reshape_fn,
-        expand_dim_fn, squeeze_fn, pad_image_fn, range_fn, round_to_int_fn,
-        repeat_fn, tile_fn, ones_like_fn, stack_fn, concat_fn,
-        image_dest_flip_lr_mat, bboxes_flip_lr_mat, matmul_fn,
-        clip_fn, gather_image_fn, copy_fn, max_fn, min_fn,
-        logical_and_fn, boolean_mask_fn
+        image, bboxes, labels, image_dest_flip_lr_mat, bboxes_flip_lr_mat,
+        interpolation, d
     )
 
 
@@ -281,27 +306,8 @@ def _flip_up_down(
         image: T,
         bboxes: T,
         labels: T,
-        convert_fn: Callable[..., T],
-        shape_fn: Callable[[T], Tuple[int, ...]],
-        reshape_fn: Callable[[T, Tuple[int, ...]], T],
-        expand_dim_fn: Callable[[T, int], T],
-        squeeze_fn: Callable[[T, int], T],
-        pad_image_fn: Callable[[T, T], T],
-        range_fn: Callable[[int, int, int], T],
-        round_to_int_fn: Callable[[T], T],
-        repeat_fn: Callable[[T, T], T],
-        tile_fn: Callable[[T, T], T],
-        ones_like_fn: Callable[[T], T],
-        stack_fn: Callable[[List[T], int], T],
-        concat_fn: Callable[[List[T], int], T],
-        matmul_fn: Callable[[T, T], T],
-        clip_fn: Callable[[T, T, T], T],
-        gather_image_fn: Callable[[T, T], T],
-        copy_fn: Callable[[T], T],
-        max_fn: Callable[[T, int], T],
-        min_fn: Callable[[T, int], T],
-        logical_and_fn: Callable[[T, T], T],
-        boolean_mask_fn: Callable[[T, T], T]
+        interpolation: Interpolation,
+        d: _AffineDependency
 ) -> Tuple[T, T, T]:
     """
     image: [h, w, c]
@@ -309,15 +315,11 @@ def _flip_up_down(
     labels: [0, 1, 0, ...]
     """
     image_dest_flip_ud_mat, bboxes_flip_ud_mat = _get_flip_up_down_mats(
-        convert_fn
+        d.convert_fn
     )
     return _affine_transform(
-        image, bboxes, labels, convert_fn, shape_fn, reshape_fn,
-        expand_dim_fn, squeeze_fn, pad_image_fn, range_fn, round_to_int_fn,
-        repeat_fn, tile_fn, ones_like_fn, stack_fn, concat_fn,
-        image_dest_flip_ud_mat, bboxes_flip_ud_mat, matmul_fn,
-        clip_fn, gather_image_fn, copy_fn, max_fn, min_fn,
-        logical_and_fn, boolean_mask_fn
+        image, bboxes, labels, image_dest_flip_ud_mat, bboxes_flip_ud_mat,
+        interpolation, d
     )
 
 
@@ -331,14 +333,14 @@ def _get_rotate_mats(
     # Image rotation matrix. Clockwise for the destination indices,
     # so the final image would appear to be rotated anti-clockwise.
     image_dest_rot_mat = convert_fn([
-        [cos_fn(ang_rad), -sin_fn(ang_rad), convert_fn(0)],
-        [sin_fn(ang_rad), cos_fn(ang_rad), convert_fn(0)],
-        [convert_fn(0), convert_fn(0), convert_fn(1)]
+        [cos_fn(ang_rad), -sin_fn(ang_rad), 0],
+        [sin_fn(ang_rad), cos_fn(ang_rad), 0],
+        [0, 0, 1]
     ])
     bboxes_rot_mat = convert_fn([  # Anti-clockwise.
-        [cos_fn(ang_rad), sin_fn(ang_rad), convert_fn(0)],
-        [-sin_fn(ang_rad), cos_fn(ang_rad), convert_fn(0)],
-        [convert_fn(0), convert_fn(0), convert_fn(1)]
+        [cos_fn(ang_rad), sin_fn(ang_rad), 0],
+        [-sin_fn(ang_rad), cos_fn(ang_rad), 0],
+        [0, 0, 1]
     ])
     return image_dest_rot_mat, bboxes_rot_mat
 
@@ -348,29 +350,10 @@ def _rotate(
         bboxes: T,
         labels: T,
         angle_deg: T,
-        convert_fn: Callable[..., T],
         cos_fn: Callable[[T], T],
         sin_fn: Callable[[T], T],
-        shape_fn: Callable[[T], Tuple[int, ...]],
-        reshape_fn: Callable[[T, Tuple[int, ...]], T],
-        expand_dim_fn: Callable[[T, int], T],
-        squeeze_fn: Callable[[T, int], T],
-        pad_image_fn: Callable[[T, T], T],
-        range_fn: Callable[[int, int, int], T],
-        round_to_int_fn: Callable[[T], T],
-        repeat_fn: Callable[[T, T], T],
-        tile_fn: Callable[[T, T], T],
-        ones_like_fn: Callable[[T], T],
-        stack_fn: Callable[[List[T], int], T],
-        concat_fn: Callable[[List[T], int], T],
-        matmul_fn: Callable[[T, T], T],
-        clip_fn: Callable[[T, T, T], T],
-        gather_image_fn: Callable[[T, T], T],
-        copy_fn: Callable[[T], T],
-        max_fn: Callable[[T, int], T],
-        min_fn: Callable[[T, int], T],
-        logical_and_fn: Callable[[T, T], T],
-        boolean_mask_fn: Callable[[T, T], T]
+        interpolation: Interpolation,
+        d: _AffineDependency
 ) -> Tuple[T, T, T]:
     """
     image: [h, w, c]
@@ -379,15 +362,11 @@ def _rotate(
     angle_deg: positive means anti-clockwise.
     """
     image_dest_rot_mat, bboxes_rot_mat = _get_rotate_mats(
-        angle_deg, convert_fn, cos_fn, sin_fn
+        angle_deg, d.convert_fn, cos_fn, sin_fn
     )
     return _affine_transform(
-        image, bboxes, labels, convert_fn, shape_fn, reshape_fn,
-        expand_dim_fn, squeeze_fn, pad_image_fn, range_fn, round_to_int_fn,
-        repeat_fn, tile_fn, ones_like_fn, stack_fn, concat_fn,
-        image_dest_rot_mat, bboxes_rot_mat, matmul_fn,
-        clip_fn, gather_image_fn, copy_fn, max_fn, min_fn,
-        logical_and_fn, boolean_mask_fn
+        image, bboxes, labels, image_dest_rot_mat, bboxes_rot_mat,
+        interpolation, d
     )
 
 
@@ -401,14 +380,14 @@ def _get_shear_mats(
     # Image shear matrix. Clockwise for the destination indices,
     # so the final image would appear to be sheared anti-clockwise.
     image_dest_shear_mat = convert_fn([
-        [convert_fn(1), -factor, convert_fn(0)],
-        [convert_fn(0), convert_fn(1), convert_fn(0)],
-        [convert_fn(0), convert_fn(0), convert_fn(1)]
+        [1, -factor, 0],
+        [0, 1, 0],
+        [0, 0, 1]
     ])
     bboxes_shear_mat = convert_fn([  # Anti-clockwise.
-        [convert_fn(1), factor, convert_fn(0)],
-        [convert_fn(0), convert_fn(1), convert_fn(0)],
-        [convert_fn(0), convert_fn(0), convert_fn(1)]
+        [1, factor, 0],
+        [0, 1, 0],
+        [0, 0, 1]
     ])
     return image_dest_shear_mat, bboxes_shear_mat
 
@@ -418,28 +397,9 @@ def _shear(
         bboxes: T,
         labels: T,
         angle_deg: T,
-        convert_fn: Callable[..., T],
         tan_fn: Callable[[T], T],
-        shape_fn: Callable[[T], Tuple[int, ...]],
-        reshape_fn: Callable[[T, Tuple[int, ...]], T],
-        expand_dim_fn: Callable[[T, int], T],
-        squeeze_fn: Callable[[T, int], T],
-        pad_image_fn: Callable[[T, T], T],
-        range_fn: Callable[[int, int, int], T],
-        round_to_int_fn: Callable[[T], T],
-        repeat_fn: Callable[[T, T], T],
-        tile_fn: Callable[[T, T], T],
-        ones_like_fn: Callable[[T], T],
-        stack_fn: Callable[[List[T], int], T],
-        concat_fn: Callable[[List[T], int], T],
-        matmul_fn: Callable[[T, T], T],
-        clip_fn: Callable[[T, T, T], T],
-        gather_image_fn: Callable[[T, T], T],
-        copy_fn: Callable[[T], T],
-        max_fn: Callable[[T, int], T],
-        min_fn: Callable[[T, int], T],
-        logical_and_fn: Callable[[T, T], T],
-        boolean_mask_fn: Callable[[T, T], T]
+        interpolation: Interpolation,
+        d: _AffineDependency
 ) -> Tuple[T, T, T]:
     """
     image: [h, w, c]
@@ -448,15 +408,11 @@ def _shear(
     angle_deg: positive means anti-clockwise, where abs(angle_deg) must be < 90.
     """
     image_dest_shear_mat, bboxes_shear_mat = _get_shear_mats(
-        angle_deg, convert_fn, tan_fn
+        angle_deg, d.convert_fn, tan_fn
     )
     return _affine_transform(
-        image, bboxes, labels, convert_fn, shape_fn, reshape_fn,
-        expand_dim_fn, squeeze_fn, pad_image_fn, range_fn, round_to_int_fn,
-        repeat_fn, tile_fn, ones_like_fn, stack_fn, concat_fn,
-        image_dest_shear_mat, bboxes_shear_mat, matmul_fn,
-        clip_fn, gather_image_fn, copy_fn, max_fn, min_fn,
-        logical_and_fn, boolean_mask_fn
+        image, bboxes, labels, image_dest_shear_mat, bboxes_shear_mat,
+        interpolation, d
     )
 
 
@@ -466,14 +422,14 @@ def _get_translate_mats(
         convert_fn: Callable[..., T]
 ) -> Tuple[T, T]:
     image_dest_translate_mat = convert_fn([
-        [convert_fn(1), convert_fn(0), -convert_fn(translate_width)],
-        [convert_fn(0), convert_fn(1), -convert_fn(translate_height)],
-        [convert_fn(0), convert_fn(0), convert_fn(1)]
+        [1, 0, -translate_width],
+        [0, 1, -translate_height],
+        [0, 0, 1]
     ])
     bboxes_translate_mat = convert_fn([
-        [convert_fn(1), convert_fn(0), convert_fn(translate_width)],
-        [convert_fn(0), convert_fn(1), convert_fn(translate_height)],
-        [convert_fn(0), convert_fn(0), convert_fn(1)]
+        [1, 0, translate_width],
+        [0, 1, translate_height],
+        [0, 0, 1]
     ])
     return image_dest_translate_mat, bboxes_translate_mat
 
@@ -484,27 +440,8 @@ def _translate(
         labels: T,
         translate_height: T,
         translate_width: T,
-        convert_fn: Callable[..., T],
-        shape_fn: Callable[[T], Tuple[int, ...]],
-        reshape_fn: Callable[[T, Tuple[int, ...]], T],
-        expand_dim_fn: Callable[[T, int], T],
-        squeeze_fn: Callable[[T, int], T],
-        pad_image_fn: Callable[[T, T], T],
-        range_fn: Callable[[int, int, int], T],
-        round_to_int_fn: Callable[[T], T],
-        repeat_fn: Callable[[T, T], T],
-        tile_fn: Callable[[T, T], T],
-        ones_like_fn: Callable[[T], T],
-        stack_fn: Callable[[List[T], int], T],
-        concat_fn: Callable[[List[T], int], T],
-        matmul_fn: Callable[[T, T], T],
-        clip_fn: Callable[[T, T, T], T],
-        gather_image_fn: Callable[[T, T], T],
-        copy_fn: Callable[[T], T],
-        max_fn: Callable[[T, int], T],
-        min_fn: Callable[[T, int], T],
-        logical_and_fn: Callable[[T, T], T],
-        boolean_mask_fn: Callable[[T, T], T]
+        interpolation: Interpolation,
+        d: _AffineDependency
 ) -> Tuple[T, T, T]:
     """
     image: [h, w, c]
@@ -514,15 +451,11 @@ def _translate(
     translate_width: in range (-image_width, image_width)
     """
     image_dest_translate_mat, bboxes_translate_mat = _get_translate_mats(
-        translate_height, translate_width, convert_fn
+        translate_height, translate_width, d.convert_fn
     )
     return _affine_transform(
-        image, bboxes, labels, convert_fn, shape_fn, reshape_fn,
-        expand_dim_fn, squeeze_fn, pad_image_fn, range_fn, round_to_int_fn,
-        repeat_fn, tile_fn, ones_like_fn, stack_fn, concat_fn,
-        image_dest_translate_mat, bboxes_translate_mat, matmul_fn,
-        clip_fn, gather_image_fn, copy_fn, max_fn, min_fn,
-        logical_and_fn, boolean_mask_fn
+        image, bboxes, labels, image_dest_translate_mat, bboxes_translate_mat,
+        interpolation, d
     )
 
 
@@ -534,7 +467,7 @@ def _get_random_size_fractions(
 ) -> Tuple[T, T]:
     """
     height_fraction_range, width_fraction_range: (-1.0, 1.0)
-    rand_fn: generate array of random numbers in range [0.0, 1.0)
+    rand_fn: generate random number in range [0.0, 1.0)
     """
     height_fraction_range = convert_fn(height_fraction_range)
     width_fraction_range = convert_fn(width_fraction_range)
@@ -544,10 +477,10 @@ def _get_random_size_fractions(
     height_fraction_diff = height_fraction_range[1] - min_height_fraction
     width_fraction_diff = width_fraction_range[1] - min_width_fraction
 
-    height_fractions = height_fraction_diff * rand_fn() + min_height_fraction
-    width_fractions = width_fraction_diff * rand_fn() + min_width_fraction
+    height_fraction = height_fraction_diff * rand_fn() + min_height_fraction
+    width_fraction = width_fraction_diff * rand_fn() + min_width_fraction
 
-    return height_fractions, width_fractions
+    return height_fraction, width_fraction
 
 
 def _get_crop_inputs(
@@ -561,29 +494,22 @@ def _get_crop_inputs(
 ) -> Tuple[T, T, T, T]:
     """
     height_fraction_range, width_fraction_range: in range [0.0, 1.0)
-    rand_fn: generate array of random numbers in range [0.0, 1.0)
+    rand_fn: generate random number in range [0.0, 1.0)
     """
     image_height = convert_fn(image_height)
     image_width = convert_fn(image_width)
 
-    height_fractions, width_fractions = _get_random_size_fractions(
+    height_fraction, width_fraction = _get_random_size_fractions(
         height_fraction_range, width_fraction_range, rand_fn, convert_fn
     )
 
-    cropped_image_heights = image_height * height_fractions
-    cropped_image_widths = image_width * width_fractions
+    crop_height = image_height * height_fraction
+    crop_width = image_width * width_fraction
 
-    offset_heights = round_to_int_fn(
-        (image_height - cropped_image_heights) * rand_fn()
-    )
-    offset_widths = round_to_int_fn(
-        (image_width - cropped_image_widths) * rand_fn()
-    )
+    offset_height = round_to_int_fn((image_height - crop_height) * rand_fn())
+    offset_width = round_to_int_fn((image_width - crop_width) * rand_fn())
 
-    return (
-        offset_heights, offset_widths,
-        cropped_image_heights, cropped_image_widths
-    )
+    return offset_height, offset_width, crop_height, crop_width
 
 
 def _crop(
@@ -592,8 +518,8 @@ def _crop(
         labels: T,
         offset_height: T,
         offset_width: T,
-        cropped_image_height: T,
-        cropped_image_width: T,
+        crop_height: T,
+        crop_width: T,
         convert_fn: Callable[..., T],
         shape_fn: Callable[[T], Tuple[int, ...]],
         reshape_fn: Callable[[T, Tuple[int, ...]], T],
@@ -616,8 +542,8 @@ def _crop(
 
     top = offset_height
     left = offset_width
-    bottom = top + cropped_image_height
-    right = left + cropped_image_width
+    bottom = top + crop_height
+    right = left + crop_width
 
     # Crop image.
     image = image[int(top):int(bottom), int(left):int(right), :]
@@ -632,15 +558,15 @@ def _crop(
     xcens = xs + widths // 2
     ycens = ys + heights // 2
     included = squeeze_fn(logical_and_fn(
-        logical_and_fn(xcens >= 0, xcens <= cropped_image_width),
-        logical_and_fn(ycens >= 0, ycens <= cropped_image_height)
+        logical_and_fn(xcens >= 0, xcens <= crop_width),
+        logical_and_fn(ycens >= 0, ycens <= crop_height)
     ), -1)  # Squeeze along the last axis.
 
     # Clip bboxes values.
-    xmaxs = clip_fn(xs + widths, convert_fn(0), cropped_image_width)
-    ymaxs = clip_fn(ys + heights, convert_fn(0), cropped_image_height)
-    xs = clip_fn(xs, convert_fn(0), cropped_image_width)
-    ys = clip_fn(ys, convert_fn(0), cropped_image_height)
+    xmaxs = clip_fn(xs + widths, convert_fn(0), crop_width)
+    ymaxs = clip_fn(ys + heights, convert_fn(0), crop_height)
+    xs = clip_fn(xs, convert_fn(0), crop_width)
+    ys = clip_fn(ys, convert_fn(0), crop_height)
     widths = xmaxs - xs
     heights = ymaxs - ys
 
@@ -668,7 +594,7 @@ def _resize(
     image: [h, w, c]
     bboxes: [[top_left_x, top_left_y, width, height], ...]
     labels: [0, 1, 0, ...]
-    dest_size: (height, width)
+    dest_size: (image_height, image_width)
     """
     image, bboxes, labels = _sanitise(
         image, bboxes, labels, convert_fn, shape_fn, reshape_fn
