@@ -5,6 +5,7 @@ from ._np import (
     shear,
     translate,
     crop,
+    resize,
     CombineAffine,
     RandomTransform,
     RandomFlipLeftRight,
@@ -13,5 +14,5 @@ from ._np import (
     RandomShear,
     RandomTranslate,
     RandomCrop,
-    Resize
+    Resize,
 )
