@@ -1,5 +1,6 @@
 from ._utils import (
+    Interpolation,
     Compose,
     collate_fn,
-    image_only
+    image_only,
 )
